/*=========================================================================

  Program:   Visualization Toolkit
  Module:    $RCSfile: vtkVolumeRayCastCompositeFunction.h,v $
  Language:  C++
  Date:      $Date: 2002/10/28 21:33:01 $
  Version:   $Revision: 1.29 $

  Copyright (c) 1993-2002 Ken Martin, Will Schroeder, Bill Lorensen 
  All rights reserved.
  See Copyright.txt or http://www.kitware.com/Copyright.htm for details.

     This software is distributed WITHOUT ANY WARRANTY; without even 
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/


// VTK-Class
// modified by Martin Haidacher



#include "VolumeRayCastCompositeFunction.h"

#include "vtkObjectFactory.h"
#include "vtkPiecewiseFunction.h"
#include "vtkVolume.h"
#include "vtkVolumeProperty.h"
#include "vtkVolumeRayCastMapper.h"

#include <math.h>


#define VTK_REMAINING_OPACITY           0.02

// This is the templated function that actually casts a ray and computes
// The composite value. This version uses nearest neighbor interpolation
// and does not perform shading.
template <class T>
void vtkCastRay_NN_Unshaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                          vtkVolumeRayCastStaticInfo *staticInfo )
{
  int             value=0;
  unsigned char   *grad_mag_ptr = NULL;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           opacity=0.0;
  float           gradient_opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  int             prev_voxel[3];
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  int             offset=0;
  int             steps_this_ray = 0;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;
  int             num_steps;
  float           *ray_start, *ray_increment;

  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;

  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();
  CTF  =  staticInfo->Volume->GetRGBArray();
  GTF  =  staticInfo->Volume->GetGrayArray();
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];
  voxel[0] = vtkRoundFuncMacro( ray_position[0] );
  voxel[1] = vtkRoundFuncMacro( ray_position[1] );
  voxel[2] = vtkRoundFuncMacro( ray_position[2] );

  // So far we haven't accumulated anything
  accum_red_intensity     = 0.0;
  accum_green_intensity   = 0.0;
  accum_blue_intensity    = 0.0;
  remaining_opacity       = 1.0;

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  
  // Keep track of previous voxel to know when we step into a new one
  // set it to something invalid to start with so that everything is
  // computed first time through
  prev_voxel[0] = voxel[0]-1;
  prev_voxel[1] = voxel[1]-1;
  prev_voxel[2] = voxel[2]-1;
  
  // Two cases - we are working with a gray or RGB transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 ) 
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Access the value at this voxel location
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0] * xinc;
        value = *(data_ptr + offset);
        opacity = SOTF[value];

        if ( opacity )
          {
          if ( grad_op_is_constant )
            {
            gradient_opacity = gradient_opacity_constant;
            }
          else 
            {
            gradient_opacity = GOTF[*(grad_mag_ptr + offset)];
            }
          opacity *= gradient_opacity;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
        
      // Accumulate some light intensity and opacity
      accum_red_intensity   += ( opacity * remaining_opacity * 
                                 GTF[(value)] );
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];
      voxel[0] = vtkRoundFuncMacro( ray_position[0] );
      voxel[1] = vtkRoundFuncMacro( ray_position[1] );
      voxel[2] = vtkRoundFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Access the value at this voxel location
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0] * xinc;
        value = *(data_ptr + offset);
        opacity = SOTF[value];

        if ( opacity )
          {
          if ( grad_op_is_constant )
            {
            gradient_opacity = gradient_opacity_constant;
            }
          else 
            {
            gradient_opacity = GOTF[*(grad_mag_ptr + offset)];  
            }
          opacity *= gradient_opacity;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }


      // Accumulate some light intensity and opacity
      accum_red_intensity   += ( opacity * remaining_opacity * 
                                 CTF[(value)*3] );
      accum_green_intensity += ( opacity * remaining_opacity * 
                                 CTF[(value)*3 + 1] );
      accum_blue_intensity  += ( opacity * remaining_opacity * 
                                 CTF[(value)*3 + 2] );
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];
      voxel[0] = vtkRoundFuncMacro( ray_position[0] );
      voxel[1] = vtkRoundFuncMacro( ray_position[1] );
      voxel[2] = vtkRoundFuncMacro( ray_position[2] );
      }
    }

  // Cap the intensity value at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }

  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;
  
}


// This is the templated function that actually casts a ray and computes
// the composite value. This version uses nearest neighbor and does
// perform shading.
template <class T>
void vtkCastRay_NN_Shaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                           vtkVolumeRayCastStaticInfo *staticInfo )
{
  int             value = 0;
  unsigned char   *grad_mag_ptr = NULL;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           opacity = 0.0;
  float           gradient_opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  int             prev_voxel[3];
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  float           *red_d_shade, *green_d_shade, *blue_d_shade;
  float           *red_s_shade, *green_s_shade, *blue_s_shade;
  unsigned short  *encoded_normals;
  float           red_shaded_value   = 0.0;
  float           green_shaded_value = 0.0;
  float           blue_shaded_value  = 0.0;
  int             offset = 0;
  int             steps_this_ray = 0;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;
  int             num_steps;
  float           *ray_start, *ray_increment;

  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;

  // Get diffuse shading table pointers
  red_d_shade = staticInfo->RedDiffuseShadingTable;
  green_d_shade = staticInfo->GreenDiffuseShadingTable;
  blue_d_shade = staticInfo->BlueDiffuseShadingTable;

  // Get specular shading table pointers
  red_s_shade = staticInfo->RedSpecularShadingTable;
  green_s_shade = staticInfo->GreenSpecularShadingTable;
  blue_s_shade = staticInfo->BlueSpecularShadingTable;

  // Get a pointer to the encoded normals for this volume
  encoded_normals = staticInfo->EncodedNormals;

  // Get the scalar opacity transfer function for this volume (which maps
  // scalar input values to opacities)
  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();

  // Get the color transfer function for this volume (which maps
  // scalar input values to RGB values)
  CTF =  staticInfo->Volume->GetRGBArray();
  GTF =  staticInfo->Volume->GetGrayArray();

  // Get the gradient opacity transfer function for this volume (which maps
  // gradient magnitudes to opacities)
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];

  voxel[0] = vtkRoundFuncMacro( ray_position[0] );
  voxel[1] = vtkRoundFuncMacro( ray_position[1] );
  voxel[2] = vtkRoundFuncMacro( ray_position[2] );

  // So far we haven't accumulated anything
  accum_red_intensity     = 0.0;
  accum_green_intensity   = 0.0;
  accum_blue_intensity    = 0.0;
  remaining_opacity       = 1.0;

  // Keep track of previous voxel to know when we step into a new one  
  prev_voxel[0] = voxel[0]-1;
  prev_voxel[1] = voxel[1]-1;
  prev_voxel[2] = voxel[2]-1;
  
  // Two cases - we are working with a gray or RGB transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 ) 
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Access the value at this voxel location and compute
      // opacity and shaded value
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0] * xinc;
        value = *(data_ptr + offset);
      
        // Get the opacity contributed by the scalar opacity transfer function
        opacity = SOTF[value];

        // Multiply by the opacity contributed by the gradient magnitude
        // transfer function (don't both if opacity is already 0)
        if ( opacity )
          {
          if ( grad_op_is_constant )
            {
            gradient_opacity = gradient_opacity_constant;
            }
          else 
            {
            gradient_opacity = GOTF[*(grad_mag_ptr + offset)];
            }
          
          opacity *= gradient_opacity;
          
          }

        // Compute the red shaded value (only if there is some opacity)
        // This is grey-scale so green and blue are the same as red
        if ( opacity )
          {
          red_shaded_value = opacity * remaining_opacity *
            ( red_d_shade[*(encoded_normals + offset)] * GTF[value] +
              red_s_shade[*(encoded_normals + offset)] );
          }
        else
          {
          red_shaded_value = 0.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
    
    
      // Accumulate the shaded intensity and opacity of this sample
      accum_red_intensity += red_shaded_value;
      remaining_opacity *= (1.0 - opacity);
    
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];
      voxel[0] = vtkRoundFuncMacro( ray_position[0] );
      voxel[1] = vtkRoundFuncMacro( ray_position[1] );
      voxel[2] = vtkRoundFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Access the value at this voxel location and compute
      // opacity and shaded value
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0] * xinc;
        value = *(data_ptr + offset);
      
        // Get the opacity contributed by the scalar opacity transfer function
        opacity = SOTF[value];

        // Multiply by the opacity contributed by the gradient magnitude
        // transfer function (don't both if opacity is already 0)
        if ( opacity )
          {
          if ( grad_op_is_constant )
            {
            gradient_opacity = gradient_opacity_constant;
            }
          else 
            {
            gradient_opacity = GOTF[*(grad_mag_ptr + offset)];
            }
          
          opacity *= gradient_opacity;    
          }     

        // Compute the red, green, and blue shaded value (only if there
        // is some opacity)
        if ( opacity )
          {
          red_shaded_value = opacity *  remaining_opacity *
            ( red_d_shade[*(encoded_normals + offset)] * CTF[value*3] +
              red_s_shade[*(encoded_normals + offset)] );
          green_shaded_value = opacity *  remaining_opacity *
            ( green_d_shade[*(encoded_normals + offset)] * CTF[value*3 + 1] +
              green_s_shade[*(encoded_normals + offset)] );
          blue_shaded_value = opacity *  remaining_opacity *
            ( blue_d_shade[*(encoded_normals + offset)] * CTF[value*3 + 2] +
              blue_s_shade[*(encoded_normals + offset)] );
          }
        else
          {
          red_shaded_value = 0.0;
          green_shaded_value = 0.0;
          blue_shaded_value = 0.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
    
    
      // Accumulate the shaded intensity and opacity of this sample
      accum_red_intensity += red_shaded_value;
      accum_green_intensity += green_shaded_value;
      accum_blue_intensity += blue_shaded_value;
      remaining_opacity *= (1.0 - opacity);
    
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];
      voxel[0] = vtkRoundFuncMacro( ray_position[0] );
      voxel[1] = vtkRoundFuncMacro( ray_position[1] );
      voxel[2] = vtkRoundFuncMacro( ray_position[2] );
      }
    }
  
  // Cap the intensities at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }
  
  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;
  
}

// This is the templated function that actually casts a ray and computes
// the composite value.  This version uses trilinear interpolation and
// does not compute shading
template <class T>
void vtkCastRay_TrilinSample_Unshaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                                       vtkVolumeRayCastStaticInfo *staticInfo, dataset datas, char *d_ptr)
{
  unsigned char   *grad_mag_ptr = NULL;
  unsigned char   *gmptr = NULL;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           red_value, green_value, blue_value;
  float           opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  float           A, B, C, D, E, F, G, H;
  float           tA, tB, tC, tD, tE, tF, tG, tH;
  int             Binc, Cinc, Dinc, Einc, Finc, Ginc, Hinc;
  T               *dptr;
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  float           x, y, z, t1, t2, t3;
  int             offset;
  float           r, g, b;
  int             steps_this_ray = 0;
  float           gradient_value;
  float           scalar_value;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;
  int             num_steps;
  float           *ray_start, *ray_increment;



  //***************

  int data_max_value = 4096;
  if (staticInfo->ScalarDataType == VTK_UNSIGNED_CHAR)
	  data_max_value = 256;

  float view_dir [3];

  float left, right, up, down, front, back;
  float gradient [3];

  float grad_length;
  float grad_dir_factor;
  float grad_factor;

  float distance;
  float foc_factor;
  float con_factor;

  float con_opacity;
  float con_color [3];

  float L1, L2, L3, L4, R1, R2, R3, R4, D1, D2, D3, D4, U1, U2, U3, U4,
		F1, F2, F3, F4, B1, B2, B3, B4;

  float LI, RE, DO, UP, FR, BA;

  int L1_inc, L2_inc, L3_inc, L4_inc, R1_inc, R2_inc, R3_inc, R4_inc,
	  D1_inc, D2_inc, D3_inc, D4_inc, U1_inc, U2_inc, U3_inc, U4_inc,
	  F1_inc, F2_inc, F3_inc, F4_inc, B1_inc, B2_inc, B3_inc, B4_inc;

  int *data_size;
  long size;

  float k_sc = datas.k_sc;
  float k_ss = datas.k_ss;
  float k_sum = k_sc + k_ss;
  k_sc /= k_sum;
  k_ss /= k_sum;
  float k_se = datas.k_se;
  float red_offset = datas.reductionOffset;
  float red_factor = datas.reductionFactor;
  float red_opacity_factor;

  float trans_factor = datas.trans_factor;;


  data_size = staticInfo->DataSize;
  size = data_size[0]*data_size[1]*data_size[2];

  char D_A;
  float d_distance = datas.d_distance, d_size = datas.d_size;
  bool dithering;
  char *dith_ptr;
  bool dithering_struct = false;
  bool dithering_context = false;
  if (datas.DITHERING && datas.DITHERING_CONTEXT)
	  dithering_context = true;

  //***************




  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;




  //*********************************

  float ray_step_length = sqrt(ray_increment[0]*ray_increment[0] + 
							   ray_increment[1]*ray_increment[1] +
							   ray_increment[2]*ray_increment[2]);

  view_dir[0] = ray_increment[0]/ray_step_length;
  view_dir[1] = ray_increment[1]/ray_step_length;
  view_dir[2] = ray_increment[2]/ray_step_length;

  //*********************************




  // Get the scalar opacity transfer function which maps scalar input values
  // to opacities
  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();

  // Get the color transfer function which maps scalar input values
  // to RGB colors
  CTF =  staticInfo->Volume->GetRGBArray();
  GTF =  staticInfo->Volume->GetGrayArray();

  // Get the gradient opacity transfer function for this volume (which maps
  // gradient magnitudes to opacities)
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];
  voxel[0] = vtkFloorFuncMacro( ray_position[0] );
  voxel[1] = vtkFloorFuncMacro( ray_position[1] );
  voxel[2] = vtkFloorFuncMacro( ray_position[2] );

  // So far we have not accumulated anything
  accum_red_intensity     = 0.0;
  accum_green_intensity   = 0.0;
  accum_blue_intensity    = 0.0;
  remaining_opacity       = 1.0;

  // Compute the increments to get to the other 7 voxel vertices from A
  Binc = xinc;
  Cinc = yinc;
  Dinc = xinc + yinc;
  Einc = zinc;
  Finc = zinc + xinc;
  Ginc = zinc + yinc;
  Hinc = zinc + xinc + yinc;




  // ***************************
  
  if ((datas.FOCUS_CONTEXT && datas.TRI_INT) || dithering_context || (datas.MAGIC_LAMP && datas.TRI_INT))
  {
	  L1_inc = -xinc;
	  L2_inc = -xinc + yinc;
	  L3_inc = -xinc + zinc;
	  L4_inc = -xinc + zinc + yinc;

	  R1_inc = 2*xinc;
	  R2_inc = 2*xinc + yinc;
	  R3_inc = 2*xinc + zinc;
	  R4_inc = 2*xinc + zinc + yinc;

	  F1_inc = -zinc;
	  F2_inc = -zinc + xinc;
	  F3_inc = -zinc + yinc;
	  F4_inc = -zinc + xinc + yinc;

	  B1_inc = 2*zinc;
	  B2_inc = 2*zinc + xinc;
	  B3_inc = 2*zinc + yinc;
	  B4_inc = 2*zinc + xinc + yinc;

	  D1_inc = -yinc;
	  D2_inc = -yinc + xinc;
	  D3_inc = -yinc + zinc;
	  D4_inc = -yinc + zinc + xinc;

	  U1_inc = 2*yinc;
	  U2_inc = 2*yinc + xinc;
	  U3_inc = 2*yinc + zinc;
	  U4_inc = 2*yinc + zinc + xinc;
  }

  //*********************************************



  
  // Two cases - we are working with a gray or RGB transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 ) 
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
      dptr = data_ptr + offset;
      
      A = *(dptr);
      B = *(dptr + Binc);
      C = *(dptr + Cinc);
      D = *(dptr + Dinc);
      E = *(dptr + Einc);
      F = *(dptr + Finc);
      G = *(dptr + Ginc);
      H = *(dptr + Hinc);
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate the value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      scalar_value = 
        A * t1 * t2 * t3 +
        B *  x * t2 * t3 +
        C * t1 *  y * t3 + 
        D *  x *  y * t3 +
        E * t1 * t2 *  z + 
        F *  x * t2 *  z + 
        G * t1 *  y *  z + 
        H *  x *  y *  z;
      
      if ( scalar_value < 0.0 ) 
        {
        scalar_value = 0.0;
        }
      else if ( scalar_value > staticInfo->Volume->GetArraySize() - 1 )
        {
        scalar_value = staticInfo->Volume->GetArraySize() - 1;
        }
      
      opacity = SOTF[(int)(scalar_value)];
      
      if ( opacity )
        {
        if ( !grad_op_is_constant )
          {
          gmptr = grad_mag_ptr + offset;
      
          A = *(gmptr);
          B = *(gmptr + Binc);
          C = *(gmptr + Cinc);
          D = *(gmptr + Dinc);
          E = *(gmptr + Einc);
          F = *(gmptr + Finc);
          G = *(gmptr + Ginc);
          H = *(gmptr + Hinc);
          
          gradient_value = 
            A * t1 * t2 * t3 +
            B *  x * t2 * t3 +
            C * t1 *  y * t3 + 
            D *  x *  y * t3 +
            E * t1 * t2 *  z + 
            F *  x * t2 *  z + 
            G * t1 *  y *  z + 
            H *  x *  y *  z;
      
          if ( gradient_value < 0.0 )
            {
            gradient_value = 0.0;
            }
          else if ( gradient_value > 255.0 )
            {
            gradient_value = 255.0;
            }

          opacity *= GOTF[(int)(gradient_value)];
          }
        else
          {
          opacity *= gradient_opacity_constant;
          }
        red_value   = opacity * GTF[(int)(scalar_value)];
        
        // Accumulate intensity and opacity for this sample location
        accum_red_intensity   += remaining_opacity * red_value;
        remaining_opacity *= (1.0 - opacity);
        }
    
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
      dptr = data_ptr + offset;
	  dith_ptr = d_ptr + offset;
      
      A = *(dptr);
      B = *(dptr + Binc);
      C = *(dptr + Cinc);
      D = *(dptr + Dinc);
      E = *(dptr + Einc);
      F = *(dptr + Finc);
      G = *(dptr + Ginc);
      H = *(dptr + Hinc);




	  //***********************************

	  if (datas.DITHERING)
	  {
		  D_A = *(dith_ptr);

		  if (D_A == 1)
			  dithering = true;
		  else
			  dithering = false;

		  if (D_A == 2)
			  dithering_struct = true;
		  else
			  dithering_struct = false;
	  }

	  if ((datas.FOCUS_CONTEXT && datas.TRI_INT) || dithering_context || (datas.MAGIC_LAMP && datas.TRI_INT))
	  {
		if ((offset%zinc)%yinc > xinc && (offset%zinc)%yinc < (yinc - 2*xinc))
		{
			L1 = *(dptr + L1_inc);
			L2 = *(dptr + L2_inc);
			L3 = *(dptr + L3_inc);
			L4 = *(dptr + L4_inc);

			R1 = *(dptr + R1_inc);
			R2 = *(dptr + R2_inc);
			R3 = *(dptr + R3_inc);
			R4 = *(dptr + R4_inc);
		}
		else
		{
			L1 = A;
			L2 = C;
			L3 = E;
			L4 = G;

			R1 = B;
			R2 = D;
			R3 = F;
			R4 = H;
		}

		if (offset > zinc && offset < (size - 2*zinc))
		{
			F1 = *(dptr + F1_inc);
			F2 = *(dptr + F2_inc);
			F3 = *(dptr + F3_inc);
			F4 = *(dptr + F4_inc);

			B1 = *(dptr + B1_inc);
			B2 = *(dptr + B2_inc);
			B3 = *(dptr + B3_inc);
			B4 = *(dptr + B4_inc);
		}
		else
		{
			F1 = A;
			F2 = B;
			F3 = C;
			F4 = D;

			B1 = E;
			B2 = F;
			B3 = G;
			B4 = H;
		}

		if (offset%zinc > yinc && offset%zinc < (zinc - 2*yinc))
		{
			D1 = *(dptr + D1_inc);
			D2 = *(dptr + D2_inc);
			D3 = *(dptr + D3_inc);
			D4 = *(dptr + D4_inc);

			U1 = *(dptr + U1_inc);
			U2 = *(dptr + U2_inc);
			U3 = *(dptr + U3_inc);
			U4 = *(dptr + U4_inc);
		}
		else
		{
			D1 = A;
			D2 = B;
			D3 = E;
			D4 = F;

			U1 = C;
			U2 = D;
			U3 = G;
			U4 = H;
		}
	  }

	  //***********************************



	  
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate the value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;

	  tA = t1 * t2 * t3;
      tB =  x * t2 * t3;
      tC = t1 *  y * t3;
      tD =  x *  y * t3;
      tE = t1 * t2 *  z;
      tF =  x * t2 *  z;
      tG = t1 *  y *  z;
      tH =  x *  y *  z;
      
      scalar_value = 
        A * tA +
        B * tB +
        C * tC + 
        D * tD +
        E * tE + 
        F * tF + 
        G * tG + 
        H * tH;

      
      if ( scalar_value < 0.0 ) 
        {
        scalar_value = 0.0;
        }
      else if ( scalar_value > staticInfo->Volume->GetArraySize() - 1 )
        {
        scalar_value = staticInfo->Volume->GetArraySize() - 1;
        }



	  // ******************************
	  // ******* eingefgt 11.03.04

	  if (datas.FOCUS_CONTEXT || dithering_context || datas.MAGIC_LAMP)
	  {
		distance = (datas.foc_middle[0]-ray_position[0])*(datas.foc_middle[0]-ray_position[0]) +
				   (datas.foc_middle[1]-ray_position[1])*(datas.foc_middle[1]-ray_position[1]) +
				   (datas.foc_middle[2]-ray_position[2])*(datas.foc_middle[2]-ray_position[2]);

		foc_factor = distance/(datas.foc_radius*datas.foc_radius);
		red_opacity_factor = 1;
		if (foc_factor > red_offset)
		{
			red_opacity_factor = pow(1/(foc_factor-red_offset+1), red_factor);
			if (red_opacity_factor < 0.2)
				red_opacity_factor = 0.2;
		}
		if (foc_factor > 1)
			foc_factor = 1;
		con_factor = pow(foc_factor, trans_factor);
		foc_factor = 1-con_factor;

		gradient[0] = 0;
		gradient[1] = 0;
		gradient[2] = 0; 


		con_opacity = datas.con_gradientTF->GetValue(scalar_value);
		grad_factor = 0;

		  
		if ((con_factor > 0 && con_opacity > 0) || (con_opacity > 0 && dithering_context) || (con_opacity > 0 && datas.MAGIC_LAMP))
		{
			if (!datas.TRI_INT)
			{
				left = A+C+E+G;
				right = B+D+F+H;
				up = C+D+G+H;
				down = A+B+E+F;
				back = E+F+G+H;
				front = A+B+C+D;

				gradient[0] = (left-right)/4;
				gradient[1] = (up-down)/4;
				gradient[2] = (front-back)/4;

				grad_length = sqrt(gradient[0]*gradient[0]+gradient[1]*gradient[1]+gradient[2]*gradient[2]);

				grad_factor = 0;
	  
				if (grad_length > 0)
				{
					gradient[0] /= grad_length;
					gradient[1] /= grad_length;
					gradient[2] /= grad_length;

					grad_dir_factor = 1 - (gradient[0]*view_dir[0] + gradient[1]*view_dir[1] + gradient[2]*view_dir[2]);
					if (grad_dir_factor > 1)
						grad_dir_factor = 2-grad_dir_factor;
					grad_factor = k_sc + k_ss*pow(grad_dir_factor, k_se);
					grad_factor *= min(data_max_value/grad_length, 1.0);
				}
			}
			else
			{
				LI = ( L1 * tA + A * tB + L2 * tC + C * tD + 
					   L3 * tE + E * tF + L4 * tG + G * tH );
				RE = ( B * tA + R1 * tB + D * tC + R2 * tD + 
					   F * tE + R3 * tF + H * tG + R4 * tH );
				FR = ( F1 * tA + F2 * tB + F3 * tC + F4 * tD + 
					   A * tE + B * tF + C * tG + D * tH );
				BA = ( E * tA + F * tB + G * tC + H * tD + 
					   B1 * tE + B2 * tF + B3 * tG + B4 * tH );
				DO = ( D1 * tA + D2 * tB + A * tC + B * tD + 
					   D3 * tE + D4 * tF + E * tG + F * tH );
				UP = ( C * tA + D * tB + U1 * tC + U2 * tD + 
					   G * tE + H * tF + U3 * tG + U4 * tH );

				gradient[0] = 0.5*LI - 0.5*RE;
				gradient[1] = 0.5*DO - 0.5*UP;
				gradient[2] = 0.5*FR - 0.5*BA;

				grad_length = sqrt(gradient[0]*gradient[0]+gradient[1]*gradient[1]+gradient[2]*gradient[2]);

				grad_factor = 0;

				if (grad_length > 0)
				{
					gradient[0] /= grad_length;
					gradient[1] /= grad_length;
					gradient[2] /= grad_length;

					grad_dir_factor = 1 - (gradient[0]*view_dir[0] + gradient[1]*view_dir[1] + gradient[2]*view_dir[2]);
					if (grad_dir_factor > 1)
						grad_dir_factor = 2-grad_dir_factor;
					grad_factor = k_sc + k_ss*pow(grad_dir_factor, k_se);
					grad_factor *= min(grad_length/data_max_value, 1.0);
				}
			}
		}
	  }

	  // *********************** ende
	  // *****************************



      
      opacity = SOTF[(int)(scalar_value)];





	  //****************************************

	  if (datas.FOCUS_CONTEXT && !datas.DITHERING)
	  {
		  opacity *= foc_factor*red_opacity_factor;
		  con_opacity *= con_factor*grad_factor*red_opacity_factor;
	  }
	  else if (!datas.FOCUS_CONTEXT && !datas.DITHERING && datas.MAGIC_LAMP)
	  {
		  con_opacity *= grad_factor;
	  }
	  else if (datas.FOCUS_CONTEXT && datas.DITHERING)
	  {
		  if (!dithering)
		  {
			  opacity = 0;
			  if (dithering_context)
				  con_opacity *= grad_factor;
			  else
				  con_opacity = 0;
		  }
		  else
		  {
			  opacity *= foc_factor*red_opacity_factor;
			  con_opacity *= con_factor*grad_factor*red_opacity_factor;
		  }
	  }
	  else if (!datas.FOCUS_CONTEXT && datas.DITHERING)
	  {
		  if (!dithering)
		  {
			  opacity = 0;
			  if (dithering_context)
				  con_opacity *= grad_factor;
			  else
				  con_opacity = 0;
		  }
		  else
			  con_opacity = 0;
	  }

	  if (datas.MAGIC_LAMP)
	  {
		  float ref_angle = gradient[0]*datas.ml_direction[0] + 
							gradient[1]*datas.ml_direction[1] + 
							gradient[2]*datas.ml_direction[2];
		  if (ref_angle > datas.ml_angle || (ref_angle < -datas.ml_angle && datas.ml_back_front))
			  con_opacity = 0;
		  else
		  {
			  if (datas.ml_transfer == 0)
				  opacity = 0;
			  else
			  {
				  float factor = 1.0f - fabs(ref_angle - datas.ml_angle)/(0.20f*datas.ml_transfer);
				  if (factor < 0)
					  factor = 0;
				  con_opacity *= (1-factor);
				  opacity *= factor;
			  }
		  }
	  }

	  //****************************************
      



      if ( opacity || con_opacity )     //**************************
        {
        if ( !grad_op_is_constant )
          {
          gmptr = grad_mag_ptr + offset;
          
          A = *(gmptr);
          B = *(gmptr + Binc);
          C = *(gmptr + Cinc);
          D = *(gmptr + Dinc);
          E = *(gmptr + Einc);
          F = *(gmptr + Finc);
          G = *(gmptr + Ginc);
          H = *(gmptr + Hinc);
          
          gradient_value = 
            A * t1 * t2 * t3 +
            B *  x * t2 * t3 +
            C * t1 *  y * t3 + 
            D *  x *  y * t3 +
            E * t1 * t2 *  z + 
            F *  x * t2 *  z + 
            G * t1 *  y *  z + 
            H *  x *  y *  z;
          
          if ( gradient_value < 0.0 )
            {
            gradient_value = 0.0;
            }
          else if ( gradient_value > 255.0 )
            {
            gradient_value = 255.0;
            }

          opacity *= GOTF[(int)(gradient_value)];
          }
        else
          {
          opacity *= gradient_opacity_constant;
          }



		r = CTF[(int)(scalar_value) * 3    ];
        g = CTF[(int)(scalar_value) * 3 + 1];
        b = CTF[(int)(scalar_value) * 3 + 2];

        red_value   = opacity * r;
        green_value = opacity * g;
        blue_value  = opacity * b;



		//*********************************

		if (datas.FOCUS_CONTEXT || ((dithering_context || dithering_struct) && !dithering) || datas.MAGIC_LAMP)
		{
			datas.con_colorTF->GetColor(scalar_value, con_color);

			if (datas.FOCUS_CONTEXT)
			{
				red_value = red_value*foc_factor + con_color[0]*con_opacity;
				green_value = green_value*foc_factor + con_color[1]*con_opacity;
				blue_value = blue_value*foc_factor + con_color[2]*con_opacity;
			}

			if (datas.MAGIC_LAMP && !datas.FOCUS_CONTEXT)
			{
				red_value = red_value + con_color[0]*con_opacity;
				green_value = green_value + con_color[1]*con_opacity;
				blue_value = blue_value + con_color[2]*con_opacity;
			}

			if (dithering_struct && datas.DITHERING)
			{
				con_opacity = datas.d_struct_opacity;

				red_value = datas.d_struct_color[0]*con_opacity;
				green_value = datas.d_struct_color[1]*con_opacity;
				blue_value = datas.d_struct_color[2]*con_opacity;
			}

			opacity += con_opacity;
			if (opacity > 1)
				opacity = 1;
		}

		//*********************************



        
        // Accumulate intensity and opacity for this sample location
        accum_red_intensity   += remaining_opacity * red_value;
        accum_green_intensity += remaining_opacity * green_value;
        accum_blue_intensity  += remaining_opacity * blue_value;
        remaining_opacity *= (1.0 - opacity);
        }
    
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    }

  // Cap the intensity value at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }

  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;


}

// This is the templated function that actually casts a ray and computes
// the composite value.  This version uses trilinear interpolation, and
// does perform shading.
template <class T>
void vtkCastRay_TrilinSample_Shaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                                     vtkVolumeRayCastStaticInfo *staticInfo, dataset datas, char *d_ptr)
{
  unsigned char   *grad_mag_ptr = NULL;
  unsigned char   *gmptr = NULL;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  float           A, B, C, D, E, F, G, H;
  int             A_n, B_n, C_n, D_n, E_n, F_n, G_n, H_n;
  float           final_rd, final_gd, final_bd;
  float           final_rs, final_gs, final_bs;
  int             Binc, Cinc, Dinc, Einc, Finc, Ginc, Hinc;
  T               *dptr;
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  float           x, y, z, t1, t2, t3;
  float           tA, tB, tC, tD, tE, tF, tG, tH;
  float           *red_d_shade, *green_d_shade, *blue_d_shade;
  float           *red_s_shade, *green_s_shade, *blue_s_shade;
  unsigned short  *encoded_normals, *nptr;
  float           red_shaded_value, green_shaded_value, blue_shaded_value;
  int             offset;
  int             steps_this_ray = 0;
  int             gradient_value;
  int             scalar_value;
  float           r, g, b;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;
  int             num_steps;
  float           *ray_start, *ray_increment;




  //***************

  int data_max_value = 4096;
  if (staticInfo->ScalarDataType == VTK_UNSIGNED_CHAR)
	  data_max_value = 256;

  float view_dir [3];

  float left, right, up, down, front, back;
  float gradient [3];

  float grad_length;
  float grad_dir_factor;
  float grad_factor;

  float distance;
  float foc_factor;
  float con_factor;

  float con_opacity;
  float con_color [3];

  float L1, L2, L3, L4, R1, R2, R3, R4, D1, D2, D3, D4, U1, U2, U3, U4,
		F1, F2, F3, F4, B1, B2, B3, B4;

  float LI, RE, DO, UP, FR, BA;

  int L1_inc, L2_inc, L3_inc, L4_inc, R1_inc, R2_inc, R3_inc, R4_inc,
	  D1_inc, D2_inc, D3_inc, D4_inc, U1_inc, U2_inc, U3_inc, U4_inc,
	  F1_inc, F2_inc, F3_inc, F4_inc, B1_inc, B2_inc, B3_inc, B4_inc;

  int *data_size;
  long size;

  float k_sc = datas.k_sc;
  float k_ss = datas.k_ss;
  float k_sum = k_sc + k_ss;
  k_sc /= k_sum;
  k_ss /= k_sum;
  float k_se = datas.k_se;
  float red_offset = datas.reductionOffset;
  float red_factor = datas.reductionFactor;
  float red_opacity_factor;

  float trans_factor = datas.trans_factor;;


  data_size = staticInfo->DataSize;
  size = data_size[0]*data_size[1]*data_size[2];

  char D_A;
  float d_distance = datas.d_distance, d_size = datas.d_size;
  bool dithering;
  char *dith_ptr;
  bool dithering_struct = false;
  bool dithering_context = false;
  if (datas.DITHERING && datas.DITHERING_CONTEXT)
	  dithering_context = true;

  //***************




  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;




  //*********************************

  float ray_step_length = sqrt(ray_increment[0]*ray_increment[0] + 
							   ray_increment[1]*ray_increment[1] +
							   ray_increment[2]*ray_increment[2]);

  view_dir[0] = ray_increment[0]/ray_step_length;
  view_dir[1] = ray_increment[1]/ray_step_length;
  view_dir[2] = ray_increment[2]/ray_step_length;

  //*********************************





  // Get diffuse shading table pointers
  red_d_shade = staticInfo->RedDiffuseShadingTable;
  green_d_shade = staticInfo->GreenDiffuseShadingTable;
  blue_d_shade = staticInfo->BlueDiffuseShadingTable;


  // Get diffuse shading table pointers
  red_s_shade = staticInfo->RedSpecularShadingTable;
  green_s_shade = staticInfo->GreenSpecularShadingTable;
  blue_s_shade = staticInfo->BlueSpecularShadingTable;

  // Get a pointer to the encoded normals for this volume
  encoded_normals = staticInfo->EncodedNormals;

  // Get the scalar opacity transfer function which maps scalar input values
  // to opacities
  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();

  // Get the color transfer function which maps scalar input values
  // to RGB values
  CTF =  staticInfo->Volume->GetRGBArray();
  GTF =  staticInfo->Volume->GetGrayArray();

  // Get the gradient opacity transfer function for this volume (which maps
  // gradient magnitudes to opacities)
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];
  voxel[0] = vtkFloorFuncMacro( ray_position[0] );
  voxel[1] = vtkFloorFuncMacro( ray_position[1] );
  voxel[2] = vtkFloorFuncMacro( ray_position[2] );

  // So far we haven't accumulated anything
  accum_red_intensity   = 0.0;
  accum_green_intensity = 0.0;
  accum_blue_intensity  = 0.0;
  remaining_opacity     = 1.0;

  // Compute the increments to get to the other 7 voxel vertices from A
  Binc = xinc;
  Cinc = yinc;
  Dinc = xinc + yinc;
  Einc = zinc;
  Finc = zinc + xinc;
  Ginc = zinc + yinc;
  Hinc = zinc + xinc + yinc;



  // ***************************
  
  if ((datas.FOCUS_CONTEXT && datas.TRI_INT) || dithering_context || (datas.MAGIC_LAMP && datas.TRI_INT))
  {
	  L1_inc = -xinc;
	  L2_inc = -xinc + yinc;
	  L3_inc = -xinc + zinc;
	  L4_inc = -xinc + zinc + yinc;

	  R1_inc = 2*xinc;
	  R2_inc = 2*xinc + yinc;
	  R3_inc = 2*xinc + zinc;
	  R4_inc = 2*xinc + zinc + yinc;

	  F1_inc = -zinc;
	  F2_inc = -zinc + xinc;
	  F3_inc = -zinc + yinc;
	  F4_inc = -zinc + xinc + yinc;

	  B1_inc = 2*zinc;
	  B2_inc = 2*zinc + xinc;
	  B3_inc = 2*zinc + yinc;
	  B4_inc = 2*zinc + xinc + yinc;

	  D1_inc = -yinc;
	  D2_inc = -yinc + xinc;
	  D3_inc = -yinc + zinc;
	  D4_inc = -yinc + zinc + xinc;

	  U1_inc = 2*yinc;
	  U2_inc = 2*yinc + xinc;
	  U3_inc = 2*yinc + zinc;
	  U4_inc = 2*yinc + zinc + xinc;
  }

  //*********************************************




  // Two cases - we are working with a gray or RGB transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 ) 
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
      dptr = data_ptr + offset;
      nptr = encoded_normals + offset;
    
      A = *(dptr);
      B = *(dptr + Binc);
      C = *(dptr + Cinc);
      D = *(dptr + Dinc);
      E = *(dptr + Einc);
      F = *(dptr + Finc);
      G = *(dptr + Ginc);
      H = *(dptr + Hinc);
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate a value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      tA = t1 * t2 * t3;
      tB =  x * t2 * t3;
      tC = t1 *  y * t3;
      tD =  x *  y * t3;
      tE = t1 * t2 *  z;
      tF =  x * t2 *  z;
      tG = t1 *  y *  z;
      tH =  x *  y *  z;
      
      scalar_value = (int) (
                            A * tA + B * tB + C * tC + D * tD + 
                            E * tE + F * tF + G * tG + H * tH );
      
      if ( scalar_value < 0 ) 
        {
        scalar_value = 0;
        }
      else if ( scalar_value > staticInfo->Volume->GetArraySize() - 1 )
        {
        scalar_value = (int)(staticInfo->Volume->GetArraySize() - 1);
        }
      
      opacity = SOTF[scalar_value];

      // If we have some opacity based on the scalar value transfer function,
      // then multiply by the opacity from the gradient magnitude transfer
      // function
      if ( opacity )
        {
        if ( !grad_op_is_constant )
          {
          gmptr = grad_mag_ptr + offset;
      
          A = *(gmptr);
          B = *(gmptr + Binc);
          C = *(gmptr + Cinc);
          D = *(gmptr + Dinc);
          E = *(gmptr + Einc);
          F = *(gmptr + Finc);
          G = *(gmptr + Ginc);
          H = *(gmptr + Hinc);
          
          gradient_value = (int) (
                                  A * tA + B * tB + C * tC + D * tD + 
                                  E * tE + F * tF + G * tG + H * tH );
          if ( gradient_value < 0 )
            {
            gradient_value = 0;
            }
          else if ( gradient_value > 255 )
            {
            gradient_value = 255;
            }
          
          opacity *= GOTF[gradient_value];
          }
        else
          {
          opacity *= gradient_opacity_constant;
          }
        }

      // If we have a combined opacity value, then compute the shading
      if ( opacity )
        {
        A_n = *(nptr);
        B_n = *(nptr + Binc);
        C_n = *(nptr + Cinc);
        D_n = *(nptr + Dinc);
        E_n = *(nptr + Einc);
        F_n = *(nptr + Finc);
        G_n = *(nptr + Ginc);
        H_n = *(nptr + Hinc);

        final_rd = 
          red_d_shade[ A_n ] * tA + red_d_shade[ B_n ] * tB +   
          red_d_shade[ C_n ] * tC + red_d_shade[ D_n ] * tD + 
          red_d_shade[ E_n ] * tE + red_d_shade[ F_n ] * tF +   
          red_d_shade[ G_n ] * tG + red_d_shade[ H_n ] * tH;
        
        final_rs = 
          red_s_shade[ A_n ] * tA + red_s_shade[ B_n ] * tB +   
          red_s_shade[ C_n ] * tC + red_s_shade[ D_n ] * tD + 
          red_s_shade[ E_n ] * tE + red_s_shade[ F_n ] * tF +   
          red_s_shade[ G_n ] * tG + red_s_shade[ H_n ] * tH;
        
        r = GTF[(scalar_value)];

        // For this sample we have do not yet have any opacity or
        // shaded intensity yet
        red_shaded_value   = opacity * ( final_rd * r + final_rs );
        
        // Accumulate intensity and opacity for this sample location   
        accum_red_intensity   += red_shaded_value * remaining_opacity;
        remaining_opacity *= (1.0 - opacity);
        }

      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
      dptr = data_ptr + offset;
      nptr = encoded_normals + offset;
	  dith_ptr = d_ptr + offset;
    
      A = *(dptr);
      B = *(dptr + Binc);
      C = *(dptr + Cinc);
      D = *(dptr + Dinc);
      E = *(dptr + Einc);
      F = *(dptr + Finc);
      G = *(dptr + Ginc);
      H = *(dptr + Hinc);




	  //***********************************

	  if (datas.DITHERING)
	  {
		  D_A = *(dith_ptr);

		  if (D_A == 1)
			  dithering = true;
		  else
			  dithering = false;

		  if (D_A == 2)
			  dithering_struct = true;
		  else
			  dithering_struct = false;
	  }

	  if ((datas.FOCUS_CONTEXT && datas.TRI_INT) || dithering_context || (datas.MAGIC_LAMP && datas.TRI_INT))
	  {
		if ((offset%zinc)%yinc > xinc && (offset%zinc)%yinc < (yinc - 2*xinc))
		{
			L1 = *(dptr + L1_inc);
			L2 = *(dptr + L2_inc);
			L3 = *(dptr + L3_inc);
			L4 = *(dptr + L4_inc);

			R1 = *(dptr + R1_inc);
			R2 = *(dptr + R2_inc);
			R3 = *(dptr + R3_inc);
			R4 = *(dptr + R4_inc);
		}
		else
		{
			L1 = A;
			L2 = C;
			L3 = E;
			L4 = G;

			R1 = B;
			R2 = D;
			R3 = F;
			R4 = H;
		}

		if (offset > zinc && offset < (size - 2*zinc))
		{
			F1 = *(dptr + F1_inc);
			F2 = *(dptr + F2_inc);
			F3 = *(dptr + F3_inc);
			F4 = *(dptr + F4_inc);

			B1 = *(dptr + B1_inc);
			B2 = *(dptr + B2_inc);
			B3 = *(dptr + B3_inc);
			B4 = *(dptr + B4_inc);
		}
		else
		{
			F1 = A;
			F2 = B;
			F3 = C;
			F4 = D;

			B1 = E;
			B2 = F;
			B3 = G;
			B4 = H;
		}

		if (offset%zinc > yinc && offset%zinc < (zinc - 2*yinc))
		{
			D1 = *(dptr + D1_inc);
			D2 = *(dptr + D2_inc);
			D3 = *(dptr + D3_inc);
			D4 = *(dptr + D4_inc);

			U1 = *(dptr + U1_inc);
			U2 = *(dptr + U2_inc);
			U3 = *(dptr + U3_inc);
			U4 = *(dptr + U4_inc);
		}
		else
		{
			D1 = A;
			D2 = B;
			D3 = E;
			D4 = F;

			U1 = C;
			U2 = D;
			U3 = G;
			U4 = H;
		}
	  }

	  //***********************************




      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate a value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;	  
      
      tA = t1 * t2 * t3;
      tB =  x * t2 * t3;
      tC = t1 *  y * t3;
      tD =  x *  y * t3;
      tE = t1 * t2 *  z;
      tF =  x * t2 *  z;
      tG = t1 *  y *  z;
      tH =  x *  y *  z;
      
      scalar_value = (int) (
                            A * tA + B * tB + C * tC + D * tD + 
                            E * tE + F * tF + G * tG + H * tH );
      
      if ( scalar_value < 0 ) 
        {
        scalar_value = 0;
        }
      else if ( scalar_value > staticInfo->Volume->GetArraySize() - 1 )
        {
        scalar_value = (int)(staticInfo->Volume->GetArraySize() - 1);
        }



	  // ******************************
	  // ******* eingefgt 11.03.04

	  if (datas.FOCUS_CONTEXT || dithering_context || datas.MAGIC_LAMP)
	  {
		distance = (datas.foc_middle[0]-ray_position[0])*(datas.foc_middle[0]-ray_position[0]) +
				   (datas.foc_middle[1]-ray_position[1])*(datas.foc_middle[1]-ray_position[1]) +
				   (datas.foc_middle[2]-ray_position[2])*(datas.foc_middle[2]-ray_position[2]);

		foc_factor = distance/(datas.foc_radius*datas.foc_radius);
		red_opacity_factor = 1;
		if (foc_factor > red_offset)
		{
			red_opacity_factor = pow(1/(foc_factor-red_offset+1), red_factor);
			if (red_opacity_factor < 0.2)
				red_opacity_factor = 0.2;
		}
		if (foc_factor > 1)
			foc_factor = 1;
		con_factor = pow(foc_factor, trans_factor);
		foc_factor = 1-con_factor;

		gradient[0] = 0;
		gradient[1] = 0;
		gradient[2] = 0; 


		con_opacity = datas.con_gradientTF->GetValue(scalar_value);
		grad_factor = 0;

		  
		if ((con_factor > 0 && con_opacity > 0) || (con_opacity > 0 && dithering_context) || (con_opacity > 0 && datas.MAGIC_LAMP))
		{
			if (!datas.TRI_INT)
			{
				left = A+C+E+G;
				right = B+D+F+H;
				up = C+D+G+H;
				down = A+B+E+F;
				back = E+F+G+H;
				front = A+B+C+D;

				gradient[0] = (left-right)/4;
				gradient[1] = (up-down)/4;
				gradient[2] = (front-back)/4;

				grad_length = sqrt(gradient[0]*gradient[0]+gradient[1]*gradient[1]+gradient[2]*gradient[2]);

				grad_factor = 0;
	  
				if (grad_length > 0)
				{
					gradient[0] /= grad_length;
					gradient[1] /= grad_length;
					gradient[2] /= grad_length;

					grad_dir_factor = 1 - (gradient[0]*view_dir[0] + gradient[1]*view_dir[1] + gradient[2]*view_dir[2]);
					if (grad_dir_factor > 1)
						grad_dir_factor = 2-grad_dir_factor;
					grad_factor = k_sc + k_ss*pow(grad_dir_factor, k_se);
					grad_factor *= min(data_max_value/grad_length, 1.0);
				}
			}
			else
			{
				LI = ( L1 * tA + A * tB + L2 * tC + C * tD + 
					   L3 * tE + E * tF + L4 * tG + G * tH );
				RE = ( B * tA + R1 * tB + D * tC + R2 * tD + 
					   F * tE + R3 * tF + H * tG + R4 * tH );
				FR = ( F1 * tA + F2 * tB + F3 * tC + F4 * tD + 
					   A * tE + B * tF + C * tG + D * tH );
				BA = ( E * tA + F * tB + G * tC + H * tD + 
					   B1 * tE + B2 * tF + B3 * tG + B4 * tH );
				DO = ( D1 * tA + D2 * tB + A * tC + B * tD + 
					   D3 * tE + D4 * tF + E * tG + F * tH );
				UP = ( C * tA + D * tB + U1 * tC + U2 * tD + 
					   G * tE + H * tF + U3 * tG + U4 * tH );

				gradient[0] = 0.5*LI - 0.5*RE;
				gradient[1] = 0.5*DO - 0.5*UP;
				gradient[2] = 0.5*FR - 0.5*BA;

				grad_length = sqrt(gradient[0]*gradient[0]+gradient[1]*gradient[1]+gradient[2]*gradient[2]);

				grad_factor = 0;

				if (grad_length > 0)
				{
					gradient[0] /= grad_length;
					gradient[1] /= grad_length;
					gradient[2] /= grad_length;

					grad_dir_factor = 1 - (gradient[0]*view_dir[0] + gradient[1]*view_dir[1] + gradient[2]*view_dir[2]);
					if (grad_dir_factor > 1)
						grad_dir_factor = 2-grad_dir_factor;
					grad_factor = k_sc + k_ss*pow(grad_dir_factor, k_se);
					grad_factor *= min(grad_length/data_max_value, 1.0);
				}
			}
		}
	  }

	  // *********************** ende
	  // *****************************


      opacity = SOTF[scalar_value];


	  //****************************************

	  if (datas.FOCUS_CONTEXT && !datas.DITHERING)
	  {
		  opacity *= foc_factor*red_opacity_factor;
		  con_opacity *= con_factor*grad_factor*red_opacity_factor;
	  }
	  else if (!datas.FOCUS_CONTEXT && !datas.DITHERING && datas.MAGIC_LAMP)
	  {
		  con_opacity *= grad_factor;
	  }
	  else if (datas.FOCUS_CONTEXT && datas.DITHERING)
	  {
		  if (!dithering)
		  {
			  opacity = 0;
			  if (dithering_context)
				  con_opacity *= grad_factor;
			  else
				  con_opacity = 0;
		  }
		  else
		  {
			  opacity *= foc_factor*red_opacity_factor;
			  con_opacity *= con_factor*grad_factor*red_opacity_factor;
		  }
	  }
	  else if (!datas.FOCUS_CONTEXT && datas.DITHERING)
	  {
		  if (!dithering)
		  {
			  opacity = 0;
			  if (dithering_context)
				  con_opacity *= grad_factor;
			  else
				  con_opacity = 0;
		  }
		  else
			  con_opacity = 0;
	  }

	  if (datas.MAGIC_LAMP)
	  {
		  float ref_angle = gradient[0]*datas.ml_direction[0] + 
							gradient[1]*datas.ml_direction[1] + 
							gradient[2]*datas.ml_direction[2];
		  if (ref_angle > datas.ml_angle || (ref_angle < -datas.ml_angle && datas.ml_back_front))
			  con_opacity = 0;
		  else
		  {
			  if (datas.ml_transfer == 0)
				  opacity = 0;
			  else
			  {
				  float factor = 1.0f - fabs(ref_angle - datas.ml_angle)/(0.20f*datas.ml_transfer);
				  if (factor < 0)
					  factor = 0;
				  con_opacity *= (1-factor);
				  opacity *= factor;
			  }
		  }
	  }

	  //****************************************

      
      if ( opacity )
	  {
          if ( !grad_op_is_constant )
            {
            gmptr = grad_mag_ptr + offset;
      
            A = *(gmptr);
            B = *(gmptr + Binc);
            C = *(gmptr + Cinc);
            D = *(gmptr + Dinc);
            E = *(gmptr + Einc);
            F = *(gmptr + Finc);
            G = *(gmptr + Ginc);
            H = *(gmptr + Hinc);
            
            gradient_value = (int) (
                                    A * tA + B * tB + C * tC + D * tD + 
                                    E * tE + F * tF + G * tG + H * tH );
            if ( gradient_value < 0 )
              {
              gradient_value = 0;
              }
            else if ( gradient_value > 255 )
              {
              gradient_value = 255;
              }

            opacity *= GOTF[gradient_value];
            }
          else
            {
            opacity *= gradient_opacity_constant;
            }
        }



      // If we have a combined opacity value, then compute the shading
      if ( opacity )
        {
        A_n = *(nptr);
        B_n = *(nptr + Binc);
        C_n = *(nptr + Cinc);
        D_n = *(nptr + Dinc);
        E_n = *(nptr + Einc);
        F_n = *(nptr + Finc);
        G_n = *(nptr + Ginc);
        H_n = *(nptr + Hinc);
        
        final_rd = 
          red_d_shade[ A_n ] * tA + red_d_shade[ B_n ] * tB +   
          red_d_shade[ C_n ] * tC + red_d_shade[ D_n ] * tD + 
          red_d_shade[ E_n ] * tE + red_d_shade[ F_n ] * tF +   
          red_d_shade[ G_n ] * tG + red_d_shade[ H_n ] * tH;
        
        final_gd = 
          green_d_shade[ A_n ] * tA + green_d_shade[ B_n ] * tB +       
          green_d_shade[ C_n ] * tC + green_d_shade[ D_n ] * tD + 
          green_d_shade[ E_n ] * tE + green_d_shade[ F_n ] * tF +       
          green_d_shade[ G_n ] * tG + green_d_shade[ H_n ] * tH;
        
        final_bd = 
          blue_d_shade[ A_n ] * tA + blue_d_shade[ B_n ] * tB +         
          blue_d_shade[ C_n ] * tC + blue_d_shade[ D_n ] * tD + 
          blue_d_shade[ E_n ] * tE + blue_d_shade[ F_n ] * tF + 
          blue_d_shade[ G_n ] * tG + blue_d_shade[ H_n ] * tH;
        
        final_rs = 
          red_s_shade[ A_n ] * tA + red_s_shade[ B_n ] * tB +   
          red_s_shade[ C_n ] * tC + red_s_shade[ D_n ] * tD + 
          red_s_shade[ E_n ] * tE + red_s_shade[ F_n ] * tF +   
          red_s_shade[ G_n ] * tG + red_s_shade[ H_n ] * tH;
        
        final_gs = 
          green_s_shade[ A_n ] * tA + green_s_shade[ B_n ] * tB +       
          green_s_shade[ C_n ] * tC + green_s_shade[ D_n ] * tD + 
          green_s_shade[ E_n ] * tE + green_s_shade[ F_n ] * tF +       
          green_s_shade[ G_n ] * tG + green_s_shade[ H_n ] * tH;
        
        final_bs = 
          blue_s_shade[ A_n ] * tA + blue_s_shade[ B_n ] * tB +         
          blue_s_shade[ C_n ] * tC + blue_s_shade[ D_n ] * tD + 
          blue_s_shade[ E_n ] * tE + blue_s_shade[ F_n ] * tF + 
          blue_s_shade[ G_n ] * tG + blue_s_shade[ H_n ] * tH;
        
        r = CTF[(scalar_value) * 3    ];
        g = CTF[(scalar_value) * 3 + 1];
        b = CTF[(scalar_value) * 3 + 2];
		
		// For this sample we have do not yet have any opacity or
        // shaded intensity yet
        red_shaded_value   = opacity * ( final_rd * r + final_rs );
        green_shaded_value = opacity * ( final_gd * g + final_gs );
        blue_shaded_value  = opacity * ( final_bd * b + final_bs );




		//*********************************

		if (datas.FOCUS_CONTEXT || (datas.MAGIC_LAMP && con_opacity > 0))
		{
			datas.con_colorTF->GetColor(scalar_value, con_color);

			red_shaded_value = red_shaded_value + con_color[0]*con_opacity;
			green_shaded_value = green_shaded_value + con_color[1]*con_opacity;
			blue_shaded_value = blue_shaded_value + con_color[2]*con_opacity;

			opacity += con_opacity;
		}

		//*********************************


        
        // Accumulate intensity and opacity for this sample location   
        accum_red_intensity   += red_shaded_value   * remaining_opacity;
        accum_green_intensity += green_shaded_value * remaining_opacity;
        accum_blue_intensity  += blue_shaded_value  * remaining_opacity;
        remaining_opacity *= (1.0 - opacity);
        }
	    //*********************************
	    else if (datas.FOCUS_CONTEXT || (!dithering && (dithering_context || dithering_struct)) || datas.MAGIC_LAMP)
		{
			datas.con_colorTF->GetColor(scalar_value, con_color);

			red_shaded_value = con_color[0]*con_opacity;
			green_shaded_value = con_color[1]*con_opacity;
			blue_shaded_value = con_color[2]*con_opacity;

			if (dithering_struct)
			{
				con_opacity = datas.d_struct_opacity;

				red_shaded_value = datas.d_struct_color[0]*con_opacity;
				green_shaded_value = datas.d_struct_color[1]*con_opacity;
				blue_shaded_value = datas.d_struct_color[2]*con_opacity;
			}

			// Accumulate intensity and opacity for this sample location   
			accum_red_intensity   += red_shaded_value   * remaining_opacity;
			accum_green_intensity += green_shaded_value * remaining_opacity;
			accum_blue_intensity  += blue_shaded_value  * remaining_opacity;
			remaining_opacity *= (1.0 - con_opacity);
		}
		//*********************************

      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    }

  // Cap the accumulated intensity at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }

  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;

 
}

// Description:
// This is the templated function that actually casts a ray and computes
// the composite value.  This version uses trilinear interpolation and
// does not compute shading
template <class T>
void vtkCastRay_TrilinVertices_Unshaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                                         vtkVolumeRayCastStaticInfo *staticInfo )
{
  unsigned char   *grad_mag_ptr = NULL;
  unsigned char   *goptr;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           red_value, green_value, blue_value;
  float           opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  int             prev_voxel[3];
  float           A, B, C, D, E, F, G, H;
  float           Ago, Bgo, Cgo, Dgo, Ego, Fgo, Ggo, Hgo;
  int             Binc, Cinc, Dinc, Einc, Finc, Ginc, Hinc;
  T               *dptr;
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  float           x, y, z, t1, t2, t3;
  float           weight;
  int             offset;
  int             steps_this_ray = 0;
  int             num_steps;
  float           *ray_start, *ray_increment;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;

  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;
 
  // Get the scalar opacity transfer function which maps scalar input values
  // to opacities
  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();

  // Get the color transfer function which maps scalar input values
  // to RGB colors
  CTF =  staticInfo->Volume->GetRGBArray();
  GTF =  staticInfo->Volume->GetGrayArray();

  // Get the gradient opacity transfer function for this volume (which maps
  // gradient magnitudes to opacities)
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];
  voxel[0] = vtkFloorFuncMacro( ray_position[0] );
  voxel[1] = vtkFloorFuncMacro( ray_position[1] );
  voxel[2] = vtkFloorFuncMacro( ray_position[2] );

  // So far we have not accumulated anything
  accum_red_intensity     = 0.0;
  accum_green_intensity   = 0.0;
  accum_blue_intensity    = 0.0;
  remaining_opacity       = 1.0;

  // Compute the increments to get to the other 7 voxel vertices from A
  Binc = xinc;
  Cinc = yinc;
  Dinc = xinc + yinc;
  Einc = zinc;
  Finc = zinc + xinc;
  Ginc = zinc + yinc;
  Hinc = zinc + xinc + yinc;
  
  // Compute the values for the first pass through the loop
  offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
  dptr   = data_ptr + offset;

  A = SOTF[(*(dptr))];
  B = SOTF[(*(dptr + Binc))];
  C = SOTF[(*(dptr + Cinc))];
  D = SOTF[(*(dptr + Dinc))];
  E = SOTF[(*(dptr + Einc))];
  F = SOTF[(*(dptr + Finc))];
  G = SOTF[(*(dptr + Ginc))];
  H = SOTF[(*(dptr + Hinc))];  

  if ( !grad_op_is_constant )
    {
    goptr  = grad_mag_ptr + offset;
    Ago = GOTF[(*(goptr))];
    Bgo = GOTF[(*(goptr + Binc))];
    Cgo = GOTF[(*(goptr + Cinc))];
    Dgo = GOTF[(*(goptr + Dinc))];
    Ego = GOTF[(*(goptr + Einc))];
    Fgo = GOTF[(*(goptr + Finc))];
    Ggo = GOTF[(*(goptr + Ginc))];
    Hgo = GOTF[(*(goptr + Hinc))];  
    }
  else 
    {
    Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
    }

  // Keep track of previous voxel to know when we step into a new one  
  prev_voxel[0] = voxel[0];
  prev_voxel[1] = voxel[1];
  prev_voxel[2] = voxel[2];

  // Two cases - we are working with a gray or RGB transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 ) 
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Have we moved into a new voxel? If so we need to recompute A-H
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
        dptr   = data_ptr + offset;
        
        A = SOTF[(*(dptr))];
        B = SOTF[(*(dptr + Binc))];
        C = SOTF[(*(dptr + Cinc))];
        D = SOTF[(*(dptr + Dinc))];
        E = SOTF[(*(dptr + Einc))];
        F = SOTF[(*(dptr + Finc))];
        G = SOTF[(*(dptr + Ginc))];
        H = SOTF[(*(dptr + Hinc))];

        if ( !grad_op_is_constant )
          {
          goptr  = grad_mag_ptr + offset;
          Ago = GOTF[(*(goptr))];
          Bgo = GOTF[(*(goptr + Binc))];
          Cgo = GOTF[(*(goptr + Cinc))];
          Dgo = GOTF[(*(goptr + Dinc))];
          Ego = GOTF[(*(goptr + Einc))];
          Fgo = GOTF[(*(goptr + Finc))];
          Ggo = GOTF[(*(goptr + Ginc))];
          Hgo = GOTF[(*(goptr + Hinc))];  
          }
        else 
          {
            Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate the value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      // For this sample we have do not yet have any opacity
      opacity          = 0.0;
      red_value        = 0.0;
      
      // Now add the opacity in vertex by vertex.  If any of the A-H
      // have a non-transparent opacity value, then add its contribution
      // to the opacity
      if ( A && Ago )
        {
        weight     = t1*t2*t3 * A * Ago;
        opacity   += weight;
        red_value += weight * GTF[((*dptr))];
        }
      
      if ( B && Bgo )
        {
        weight     = x*t2*t3 * B * Bgo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Binc)))];
        }
      
      if ( C && Cgo )
        {
        weight     = t1*y*t3 * C * Cgo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Cinc)))];
        }
      
      if ( D && Dgo )
        {
        weight     = x*y*t3 * D * Dgo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Dinc)))];
        }
      
      if ( E && Ego )
        {
        weight     = t1*t2*z * E * Ego; 
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Einc)))];
        }
      
      if ( F && Fgo )
        {
        weight     = x*t2*z * F * Fgo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Finc)))];
        }
      
      if ( G && Ggo )
        {
        weight     = t1*y*z * G * Ggo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Ginc)))];
        } 
      
      if ( H && Hgo )
        {
        weight     = x*z*y * H * Hgo;
        opacity   += weight;
        red_value += weight * GTF[((*(dptr + Hinc)))];
        }
      
      // Accumulate intensity and opacity for this sample location
      accum_red_intensity   += remaining_opacity * red_value;
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Have we moved into a new voxel? If so we need to recompute A-H
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
        dptr = data_ptr + offset;
        
        A = SOTF[(*(dptr))];
        B = SOTF[(*(dptr + Binc))];
        C = SOTF[(*(dptr + Cinc))];
        D = SOTF[(*(dptr + Dinc))];
        E = SOTF[(*(dptr + Einc))];
        F = SOTF[(*(dptr + Finc))];
        G = SOTF[(*(dptr + Ginc))];
        H = SOTF[(*(dptr + Hinc))];

        if ( !grad_op_is_constant ) 
          {
          goptr  = grad_mag_ptr + offset;
          Ago = GOTF[(*(goptr))];
          Bgo = GOTF[(*(goptr + Binc))];
          Cgo = GOTF[(*(goptr + Cinc))];
          Dgo = GOTF[(*(goptr + Dinc))];
          Ego = GOTF[(*(goptr + Einc))];
          Fgo = GOTF[(*(goptr + Finc))];
          Ggo = GOTF[(*(goptr + Ginc))];
          Hgo = GOTF[(*(goptr + Hinc))];  
          }
        else 
          {
            Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate the value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      // For this sample we have do not yet have any opacity
      opacity           = 0.0;
      red_value         = 0.0;
      green_value       = 0.0;
      blue_value        = 0.0;
      
      // Now add the opacity in vertex by vertex.  If any of the A-H
      // have a non-transparent opacity value, then add its contribution
      // to the opacity
      if ( A && Ago )
        {
        weight         = t1*t2*t3 * A * Ago;
        opacity       += weight;
        red_value     += weight * CTF[((*dptr)) * 3    ];
        green_value   += weight * CTF[((*dptr)) * 3 + 1];
        blue_value    += weight * CTF[((*dptr)) * 3 + 2];
        }
      
      if ( B && Bgo )
        {
        weight         = x*t2*t3 * B * Bgo;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Binc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Binc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Binc))) * 3 + 2];
        }
      
      if ( C && Cgo )
        {
        weight         = t1*y*t3 * C * Cgo;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Cinc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Cinc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Cinc))) * 3 + 2];
        }
      
      if ( D && Dgo )
        {
        weight         = x*y*t3 * D * Dgo;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Dinc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Dinc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Dinc))) * 3 + 2];
        }
      
      if ( E && Ego )
        {
        weight         = t1*t2*z * E * Ego;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Einc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Einc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Einc))) * 3 + 2];
        }
      
      if ( F && Fgo )
        {
        weight         = x*t2*z * F * Fgo;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Finc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Finc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Finc))) * 3 + 2];
        }
      
      if ( G && Ggo )
        {
        weight         = t1*y*z * G * Ggo;
        opacity       += weight;
        red_value     +=  weight * CTF[((*(dptr + Ginc))) * 3    ];
        green_value   +=  weight * CTF[((*(dptr + Ginc))) * 3 + 1];
        blue_value    +=  weight * CTF[((*(dptr + Ginc))) * 3 + 2];
        } 
      
      if ( H && Hgo )
        {
        weight         = x*z*y * H * Hgo;
        opacity       += weight;
        red_value     += weight * CTF[((*(dptr + Hinc))) * 3    ];
        green_value   += weight * CTF[((*(dptr + Hinc))) * 3 + 1];
        blue_value    += weight * CTF[((*(dptr + Hinc))) * 3 + 2];
        }
      
      // Accumulate intensity and opacity for this sample location
      accum_red_intensity   += remaining_opacity * red_value;
      accum_green_intensity += remaining_opacity * green_value;
      accum_blue_intensity  += remaining_opacity * blue_value;
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    }

  // Cap the accumulated intensity at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }

  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;


}


// Description:
// This is the templated function that actually casts a ray and computes
// the composite value.  This version uses trilinear interpolation, and
// does perform shading.
template <class T>
void vtkCastRay_TrilinVertices_Shaded( T *data_ptr, vtkVolumeRayCastDynamicInfo *dynamicInfo,
                                       vtkVolumeRayCastStaticInfo *staticInfo )
{
  unsigned char   *grad_mag_ptr = NULL;
  unsigned char   *goptr;
  float           accum_red_intensity;
  float           accum_green_intensity;
  float           accum_blue_intensity;
  float           remaining_opacity;
  float           opacity;
  int             loop;
  int             xinc, yinc, zinc;
  int             voxel[3];
  float           ray_position[3];
  int             prev_voxel[3];
  float           A, B, C, D, E, F, G, H;
  float           Ago, Bgo, Cgo, Dgo, Ego, Fgo, Ggo, Hgo;
  int             Binc, Cinc, Dinc, Einc, Finc, Ginc, Hinc;
  T               *dptr;
  float           *SOTF;
  float           *CTF;
  float           *GTF;
  float           *GOTF;
  float           x, y, z, t1, t2, t3;
  float           weight;
  float           *red_d_shade, *green_d_shade, *blue_d_shade;
  float           *red_s_shade, *green_s_shade, *blue_s_shade;
  unsigned short  *encoded_normals, *nptr;
  float           red_shaded_value, green_shaded_value, blue_shaded_value;
  int             offset;
  int             steps_this_ray = 0;
  int             grad_op_is_constant;
  float           gradient_opacity_constant;
  int             num_steps;
  float           *ray_start, *ray_increment;

  num_steps = dynamicInfo->NumberOfStepsToTake;
  ray_start = dynamicInfo->TransformedStart;
  ray_increment = dynamicInfo->TransformedIncrement;

  // Get diffuse shading table pointers
  red_d_shade = staticInfo->RedDiffuseShadingTable;
  green_d_shade = staticInfo->GreenDiffuseShadingTable;
  blue_d_shade = staticInfo->BlueDiffuseShadingTable;


  // Get diffuse shading table pointers
  red_s_shade = staticInfo->RedSpecularShadingTable;
  green_s_shade = staticInfo->GreenSpecularShadingTable;
  blue_s_shade = staticInfo->BlueSpecularShadingTable;

  // Get a pointer to the encoded normals for this volume
  encoded_normals = staticInfo->EncodedNormals;

  // Get the scalar opacity transfer function which maps scalar input values
  // to opacities
  SOTF =  staticInfo->Volume->GetCorrectedScalarOpacityArray();

  // Get the color transfer function which maps scalar input values
  // to RGB values
  CTF =  staticInfo->Volume->GetRGBArray();
  GTF =  staticInfo->Volume->GetGrayArray();

  // Get the gradient opacity transfer function for this volume (which maps
  // gradient magnitudes to opacities)
  GOTF =  staticInfo->Volume->GetGradientOpacityArray();

  // Get the gradient opacity constant. If this number is greater than
  // or equal to 0.0, then the gradient opacity transfer function is
  // a constant at that value, otherwise it is not a constant function
  gradient_opacity_constant = staticInfo->Volume->GetGradientOpacityConstant();
  grad_op_is_constant = ( gradient_opacity_constant >= 0.0 );

  // Get a pointer to the gradient magnitudes for this volume
  if ( !grad_op_is_constant )
    {
    grad_mag_ptr = staticInfo->GradientMagnitudes;
    }

  // Move the increments into local variables
  xinc = staticInfo->DataIncrement[0];
  yinc = staticInfo->DataIncrement[1];
  zinc = staticInfo->DataIncrement[2];

  // Initialize the ray position and voxel location
  ray_position[0] = ray_start[0];
  ray_position[1] = ray_start[1];
  ray_position[2] = ray_start[2];
  voxel[0] = vtkFloorFuncMacro( ray_position[0] );
  voxel[1] = vtkFloorFuncMacro( ray_position[1] );
  voxel[2] = vtkFloorFuncMacro( ray_position[2] );

  // So far we haven't accumulated anything
  accum_red_intensity     = 0.0;
  accum_green_intensity   = 0.0;
  accum_blue_intensity    = 0.0;
  remaining_opacity       = 1.0;

  // Compute the increments to get to the other 7 voxel vertices from A
  Binc = xinc;
  Cinc = yinc;
  Dinc = xinc + yinc;
  Einc = zinc;
  Finc = zinc + xinc;
  Ginc = zinc + yinc;
  Hinc = zinc + xinc + yinc;
  
   // Compute the values for the first pass through the loop
  offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
  dptr   = data_ptr + offset;
  nptr   = encoded_normals + offset;

  A = SOTF[(*(dptr))];
  B = SOTF[(*(dptr + Binc))];
  C = SOTF[(*(dptr + Cinc))];
  D = SOTF[(*(dptr + Dinc))];
  E = SOTF[(*(dptr + Einc))];
  F = SOTF[(*(dptr + Finc))];
  G = SOTF[(*(dptr + Ginc))];
  H = SOTF[(*(dptr + Hinc))];

  if ( !grad_op_is_constant )
    {
    goptr  = grad_mag_ptr + offset;
    Ago = GOTF[(*(goptr))];
    Bgo = GOTF[(*(goptr + Binc))];
    Cgo = GOTF[(*(goptr + Cinc))];
    Dgo = GOTF[(*(goptr + Dinc))];
    Ego = GOTF[(*(goptr + Einc))];
    Fgo = GOTF[(*(goptr + Finc))];
    Ggo = GOTF[(*(goptr + Ginc))];
    Hgo = GOTF[(*(goptr + Hinc))];  
    }
  else
    {
    Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
    }


  // Keep track of previous voxel to know when we step into a new one   
  prev_voxel[0] = voxel[0];
  prev_voxel[1] = voxel[1];
  prev_voxel[2] = voxel[2];

  // Two cases - we are working with a single color or a color transfer
  // function - break them up to make it more efficient
  if ( staticInfo->ColorChannels == 1 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Have we moved into a new voxel? If so we need to recompute A-H
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
        dptr   = data_ptr + offset;
        nptr   = encoded_normals + offset;
        goptr  = grad_mag_ptr + offset;
        
        A = SOTF[(*(dptr))];
        B = SOTF[(*(dptr + Binc))];
        C = SOTF[(*(dptr + Cinc))];
        D = SOTF[(*(dptr + Dinc))];
        E = SOTF[(*(dptr + Einc))];
        F = SOTF[(*(dptr + Finc))];
        G = SOTF[(*(dptr + Ginc))];
        H = SOTF[(*(dptr + Hinc))];
        
        if ( !grad_op_is_constant )
          {
          goptr  = grad_mag_ptr + offset;
          Ago = GOTF[(*(goptr))];
          Bgo = GOTF[(*(goptr + Binc))];
          Cgo = GOTF[(*(goptr + Cinc))];
          Dgo = GOTF[(*(goptr + Dinc))];
          Ego = GOTF[(*(goptr + Einc))];
          Fgo = GOTF[(*(goptr + Finc))];
          Ggo = GOTF[(*(goptr + Ginc))];
          Hgo = GOTF[(*(goptr + Hinc))];  
          }
        else
          {
          Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate a value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      // For this sample we have do not yet have any opacity or
      // shaded intensity yet
      opacity = 0.0;
      red_shaded_value = 0.0;
      
      // Now add the opacity and shaded intensity value in vertex by 
      // vertex.  If any of the A-H have a non-transparent opacity value, 
      // then add its contribution to the opacity
      if ( A && Ago )
        {
        weight = t1*t2*t3 * A * Ago;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr) ] * 
                                     GTF[*(dptr)] + 
                                     red_s_shade[ *(nptr) ] );
        }
      
      if ( B && Bgo )
        {
        weight = x*t2*t3 * B * Bgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Binc) ] * 
                                     GTF[*(dptr+Binc)] + 
                                     red_s_shade[ *(nptr + Binc) ] );
        }
      
      if ( C && Cgo )
        {
        weight = t1*y*t3 * C * Cgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Cinc) ] *
                                     GTF[*(dptr+Cinc)] + 
                                     red_s_shade[ *(nptr + Cinc) ] );
        }

      if ( D && Dgo )
        {
        weight = x*y*t3 * D * Dgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Dinc) ] *
                                     GTF[*(dptr+Dinc)] + 
                                     red_s_shade[ *(nptr + Dinc) ] );
        }
      
      if ( E && Ego )
        {
        weight = t1*t2*z * E * Ego;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Einc) ] *
                                     GTF[*(dptr+Einc)] + 
                                     red_s_shade[ *(nptr + Einc) ] );
        }
      
      if ( F && Fgo )
        {
        weight = x*z*t2 * F * Fgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Finc) ] *
                                     GTF[*(dptr+Finc)] + 
                                     red_s_shade[ *(nptr + Finc) ] );
        }
      
      if ( G && Ggo )
        {
        weight = t1*y*z * G * Ggo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Ginc) ] *
                                     GTF[*(dptr+Ginc)] + 
                                     red_s_shade[ *(nptr + Ginc) ] );
      } 
      
      if ( H && Hgo )
        {
        weight = x*z*y * H * Hgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Hinc) ] *
                                     GTF[*(dptr+Hinc)] + 
                                     red_s_shade[ *(nptr + Hinc) ] );
        }
      
      // Accumulate intensity and opacity for this sample location   
      accum_red_intensity   += red_shaded_value   * remaining_opacity;
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    accum_green_intensity = accum_red_intensity;
    accum_blue_intensity = accum_red_intensity;
    }
  else if ( staticInfo->ColorChannels == 3 )
    {
    // For each step along the ray
    for ( loop = 0; 
          loop < num_steps && remaining_opacity > VTK_REMAINING_OPACITY; 
          loop++ )
      {     
      // We've taken another step
      steps_this_ray++;
      
      // Have we moved into a new voxel? If so we need to recompute A-H
      if ( prev_voxel[0] != voxel[0] ||
           prev_voxel[1] != voxel[1] ||
           prev_voxel[2] != voxel[2] )
        {
        offset = voxel[2] * zinc + voxel[1] * yinc + voxel[0];
        dptr   = data_ptr + offset;
        nptr   = encoded_normals + offset;
        goptr  = grad_mag_ptr + offset;
        
        A = SOTF[(*(dptr))];
        B = SOTF[(*(dptr + Binc))];
        C = SOTF[(*(dptr + Cinc))];
        D = SOTF[(*(dptr + Dinc))];
        E = SOTF[(*(dptr + Einc))];
        F = SOTF[(*(dptr + Finc))];
        G = SOTF[(*(dptr + Ginc))];
        H = SOTF[(*(dptr + Hinc))];
        
        if ( !grad_op_is_constant )
          {
          goptr  = grad_mag_ptr + offset;
          Ago = GOTF[(*(goptr))];
          Bgo = GOTF[(*(goptr + Binc))];
          Cgo = GOTF[(*(goptr + Cinc))];
          Dgo = GOTF[(*(goptr + Dinc))];
          Ego = GOTF[(*(goptr + Einc))];
          Fgo = GOTF[(*(goptr + Finc))];
          Ggo = GOTF[(*(goptr + Ginc))];
          Hgo = GOTF[(*(goptr + Hinc))];  
          }
        else
          {
          Ago = Bgo = Cgo = Dgo = Ego = Fgo = Ggo = Hgo = 1.0;
          }

        prev_voxel[0] = voxel[0];
        prev_voxel[1] = voxel[1];
        prev_voxel[2] = voxel[2];
        }
      
      // Compute our offset in the voxel, and use that to trilinearly
      // interpolate a value
      x = ray_position[0] - (float) voxel[0];
      y = ray_position[1] - (float) voxel[1];
      z = ray_position[2] - (float) voxel[2];
      
      t1 = 1.0 - x;
      t2 = 1.0 - y;
      t3 = 1.0 - z;
      
      // For this sample we have do not yet have any opacity or
      // shaded intensity yet
      opacity = 0.0;
      red_shaded_value = 0.0;
      green_shaded_value = 0.0;
      blue_shaded_value = 0.0;
      
      // Now add the opacity and shaded intensity value in vertex by 
      // vertex.  If any of the A-H have a non-transparent opacity value, 
      // then add its contribution to the opacity
      if ( A )
        {
        weight = t1*t2*t3 * A * Ago;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr) ] * 
                                     CTF[*(dptr) * 3] + 
                                     red_s_shade[ *(nptr) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr) ] * 
                                     CTF[*(dptr) * 3 + 1] + + 
                                     green_s_shade[ *(nptr) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr) ] * 
                                     CTF[*(dptr) * 3 + 2] +
                                     blue_s_shade[ *(nptr) ]  );
        }
      
      if ( B )
        {
        weight = x*t2*t3 * B * Bgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Binc) ] * 
                                     CTF[*(dptr+Binc) * 3] + 
                                     red_s_shade[ *(nptr + Binc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Binc) ] *
                                     CTF[*(dptr+Binc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Binc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Binc) ] *
                                     CTF[*(dptr+Binc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Binc) ] );
        }
      
      if ( C )
        {
        weight = t1*y*t3 * C * Cgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Cinc) ] *
                                     CTF[*(dptr+Cinc) * 3] + 
                                     red_s_shade[ *(nptr + Cinc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Cinc) ] *
                                     CTF[*(dptr+Cinc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Cinc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Cinc) ] *
                                     CTF[*(dptr+Cinc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Cinc) ] );
        }

      if ( D )
        {
        weight = x*y*t3 * D * Dgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Dinc) ] *
                                     CTF[*(dptr+Dinc) * 3] + 
                                     red_s_shade[ *(nptr + Dinc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Dinc) ] *
                                     CTF[*(dptr+Dinc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Dinc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Dinc) ] *
                                     CTF[*(dptr+Dinc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Dinc) ] );
        }
      
      if ( E )
        {
        weight = t1*t2*z * E * Ego;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Einc) ] *
                                     CTF[*(dptr+Einc) * 3] + 
                                     red_s_shade[ *(nptr + Einc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Einc) ] *
                                     CTF[*(dptr+Einc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Einc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Einc) ] *
                                     CTF[*(dptr+Einc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Einc) ] );
        }
      
      if ( F )
        {
        weight = x*z*t2 * F * Fgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Finc) ] *
                                     CTF[*(dptr+Finc) * 3] + 
                                     red_s_shade[ *(nptr + Finc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Finc) ] *
                                     CTF[*(dptr+Finc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Finc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Finc) ] *
                                     CTF[*(dptr+Finc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Finc) ] );
        }
      
      if ( G )
        {
        weight = t1*y*z * G * Ggo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Ginc) ] *
                                     CTF[*(dptr+Ginc) * 3] + 
                                     red_s_shade[ *(nptr + Ginc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Ginc) ] *
                                     CTF[*(dptr+Ginc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Ginc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Ginc) ] *
                                     CTF[*(dptr+Ginc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Ginc) ] );
      } 
      
      if ( H )
        {
        weight = x*z*y * H * Hgo;
        opacity += weight;
        red_shaded_value   += weight * ( red_d_shade[ *(nptr + Hinc) ] *
                                     CTF[*(dptr+Hinc) * 3] + 
                                     red_s_shade[ *(nptr + Hinc) ] );
        green_shaded_value += weight * ( green_d_shade[ *(nptr + Hinc) ] *
                                     CTF[*(dptr+Hinc) * 3 + 1] + 
                                     green_s_shade[ *(nptr + Hinc) ] );
        blue_shaded_value  += weight * ( blue_d_shade[ *(nptr + Hinc) ] *
                                     CTF[*(dptr+Hinc) * 3 + 2] +
                                     blue_s_shade[ *(nptr + Hinc) ] );
        }
      
      // Accumulate intensity and opacity for this sample location   
      accum_red_intensity   += red_shaded_value   * remaining_opacity;
      accum_green_intensity += green_shaded_value * remaining_opacity;
      accum_blue_intensity  += blue_shaded_value  * remaining_opacity;
      remaining_opacity *= (1.0 - opacity);
      
      // Increment our position and compute our voxel location
      ray_position[0] += ray_increment[0];
      ray_position[1] += ray_increment[1];
      ray_position[2] += ray_increment[2];      
      voxel[0] = vtkFloorFuncMacro( ray_position[0] );
      voxel[1] = vtkFloorFuncMacro( ray_position[1] );
      voxel[2] = vtkFloorFuncMacro( ray_position[2] );
      }
    }

  // Cap the accumulated intensity at 1.0
  if ( accum_red_intensity > 1.0 )
    {
    accum_red_intensity = 1.0;
    }
  if ( accum_green_intensity > 1.0 )
    {
    accum_green_intensity = 1.0;
    }
  if ( accum_blue_intensity > 1.0 )
    {
    accum_blue_intensity = 1.0;
    }
  
  if( remaining_opacity < VTK_REMAINING_OPACITY )
    {
    remaining_opacity = 0.0;
    }

  // Set the return pixel value.  The depth value is the distance to the
  // center of the volume.
  dynamicInfo->Color[0] = accum_red_intensity;
  dynamicInfo->Color[1] = accum_green_intensity;
  dynamicInfo->Color[2] = accum_blue_intensity;
  dynamicInfo->Color[3] = 1.0 - remaining_opacity;
  dynamicInfo->NumberOfStepsTaken = steps_this_ray;


}

// Constructor for the vtkVolumeRayCastCompositeFunction class
VolumeRayCastCompositeFunction::VolumeRayCastCompositeFunction()
{
  this->CompositeMethod = VTK_COMPOSITE_INTERPOLATE_FIRST;
  datas.foc_middle = new float [3];
  datas.con_colorTF = vtkColorTransferFunction::New();
  datas.con_gradientTF = vtkPiecewiseFunction::New();
  datas.foc_radius = 64;
  datas.foc_middle[0] = 128;
  datas.foc_middle[1] = 128;
  datas.foc_middle[2] = 128;

  datas.FOCUS_CONTEXT = false;
  context_volume_loaded = false;
  datas.DITHERING = false;
  datas.DITHERING_CONTEXT = true;
  datas.DITHERING_STRUCT = true;
  datas.DITHERING_SECROW = true;
  datas.DITHERING_SECROW_STRUCT = false;
  dithering_pattern_loaded = false;
  datas.MAGIC_LAMP = false;
  datas.TRI_INT = true;
  datas.k_sc = 0.05;
  datas.k_ss = 0.95;
  datas.k_se = 0.05;
  datas.trans_factor = 2;
  datas.reductionFactor = 1;
  datas.reductionOffset = 1;
  datas.d_distance = 1;
  datas.d_size = 1;
  datas.d_struct_color[0] = 1;
  datas.d_struct_color[1] = 1;
  datas.d_struct_color[2] = 1;
  datas.d_struct_opacity = 0.1;
  datas.ml_direction[0] = 0;
  datas.ml_direction[1] = 0;
  datas.ml_direction[2] = 1;
  datas.ml_angle = cos(15*(3.14159265)/180);
  datas.ml_back_front = false;
  datas.ml_transfer = 0;
}

// Destruct the vtkVolumeRayCastCompositeFunction
VolumeRayCastCompositeFunction::~VolumeRayCastCompositeFunction()
{
	delete ditheringPattern;
}


void VolumeRayCastCompositeFunction::setFocusContextEnable()
{
	if (context_volume_loaded)
		datas.FOCUS_CONTEXT = true;
	else
		vtkWarningMacro ( << "No Volume for context is being set." );
}


void VolumeRayCastCompositeFunction::setFocusContextDisable()
{
	datas.FOCUS_CONTEXT = false;
}


bool VolumeRayCastCompositeFunction::getFocusContextStatus()
{
	return datas.FOCUS_CONTEXT;
}


void VolumeRayCastCompositeFunction::setFocusArea(float *middle, float radius)
{
	datas.foc_radius = radius;
	datas.foc_middle[0] = middle[0];
	datas.foc_middle[1] = middle[1];
	datas.foc_middle[2] = middle[2];
}


void VolumeRayCastCompositeFunction::setContextVolume(vtkColorTransferFunction *con_colorTF, vtkPiecewiseFunction * con_gradientTF)
{
	datas.con_colorTF = con_colorTF;
	datas.con_gradientTF = con_gradientTF;
	context_volume_loaded = true;
}


void VolumeRayCastCompositeFunction::setParameters(float transC, float ksc, float kss, float kse, float reducOffset, float reducFac)
{
	datas.trans_factor = transC;
	datas.k_sc = ksc;
	datas.k_ss = kss;
	datas.k_se = kse;
	datas.reductionFactor = reducFac;
	datas.reductionOffset = reducOffset;
}


void VolumeRayCastCompositeFunction::setDithering(int dim [3], float spacing [3], float distance, float size, bool context, bool structure, bool second_row, bool second_row_structure)
{
	delete ditheringPattern;
	datas.DITHERING_CONTEXT = context;
	datas.DITHERING_STRUCT = structure;
	datas.DITHERING_SECROW = second_row;
	datas.DITHERING_SECROW_STRUCT = second_row_structure;
	ditheringPattern = new char [dim[0]*dim[1]*dim[2]];
	long i;
	for (i = 0; i < dim[0]*dim[1]*dim[2]; i++)
		ditheringPattern[i] = 0;
	setPoints(ditheringPattern, dim, spacing, distance, size);
	dithering_pattern_loaded = true;
}


void VolumeRayCastCompositeFunction::setDithering(bool context, bool structure, bool second_row, bool second_row_structure)
{
	datas.DITHERING_CONTEXT = context;
	datas.DITHERING_STRUCT = structure;
	datas.DITHERING_SECROW = second_row;
	datas.DITHERING_SECROW_STRUCT = second_row_structure;
}


void VolumeRayCastCompositeFunction::setPoints(char *array, int dim [3], float spacing [3], float distance_rel, float size_rel)
{
	float help = max(dim[0], max(dim[1], dim[2]))/2;
	bool flag = true;
	float size [3];
	float min_spacing = min(spacing[0], min(spacing[1], spacing[2]));
	if (min_spacing < 1)
	{
		spacing[0] /= min_spacing;
		spacing[1] /= min_spacing;
		spacing[2] /= min_spacing;
	}
	spacing[0] = 1/spacing[0];
	spacing[1] = 1/spacing[1];
	spacing[2] = 1/spacing[2];
	size[0] = floor(size_rel*(log(help)/log(2))*spacing[0]);
	size[1] = floor(size_rel*(log(help)/log(2))*spacing[1]);
	size[2] = floor(size_rel*(log(help)/log(2))*spacing[2]);
	float step [3];
	step[0] = (2*(size[0]+1) + (size[0]+1)*distance_rel);
	step[1] = (2*(size[1]+1) + (size[1]+1)*distance_rel);
	step[2] = (2*(size[2]+1) + (size[2]+1)*distance_rel);
	int count [3];
	count[0] = (dim[0]/step[0])+1;
	count[1] = (dim[1]/step[1])+1;
	count[2] = (dim[2]/step[2])+1;
	int point [3];
	point[0] = size[0]/2;
	for (int i = 0; i < count[0]; i++)
	{
		point[1] = size[1]/2;
		for (int j = 0; j < count[1]; j++)
		{
			point[2] = size[2]/2;
			for (int k = 0; k < count[2]; k++)
			{
				setRegion(array, dim, point, size, 1);
				point[2] += step[2];
			}
			point[1] += step[1];
		}
		point[0] += step[0];
	}


	if (datas.DITHERING_STRUCT)
	{
		count[0] = (dim[0]/step[0])+1;
		count[1] = (dim[1]/step[1])+1;
		count[2] = (dim[2]/step[2])+1;
		point[0] = size[0]/2;
		for (i = 0; i < count[0]; i++)
		{
			point[1] = size[1]/2;
			for (int j = 0; j < count[1]; j++)
			{
				point[2] = size[2]/2;
				for (int k = 0; k < count[2]; k++)
				{
					setStruct(array, dim, point, step, 2);
					point[2] += step[2];
				}
				point[1] += step[1];
			}
			point[0] += step[0];
		}
	}

	if (datas.DITHERING_SECROW)
	{
		count[0] = ((dim[0]-size[0]-1)/step[0])+1;
		count[1] = ((dim[1]-size[1]-1)/step[1])+1;
		count[2] = ((dim[2]-size[2]-1)/step[2])+1;
		point[0] = size[0]+1 + size[0]/2 + (size[0]+1)*distance_rel/2;
		for (i = 0; i < count[0]; i++)
		{
			point[1] = size[1]+1 + size[1]/2 + (size[1]+1)*distance_rel/2;
			for (int j = 0; j < count[1]; j++)
			{
				point[2] = size[2]+1 + size[2]/2 + (size[2]+1)*distance_rel/2;
				for (int k = 0; k < count[2]; k++)
				{
					setRegion(array, dim, point, size, 1);
					point[2] += step[2];
				}
				point[1] += step[1];
			}
			point[0] += step[0];
		}


		if (datas.DITHERING_SECROW_STRUCT)
		{
			count[0] = ((dim[0]-size[0]-1)/step[0]);
			count[1] = ((dim[1]-size[1]-1)/step[1]);
			count[2] = ((dim[2]-size[2]-1)/step[2]);
			point[0] = size[0]+1 + size[0]/2 + (size[0]+1)*distance_rel/2;
			for (i = 0; i < count[0]; i++)
			{
				point[1] = size[1]+1 + size[1]/2 + (size[1]+1)*distance_rel/2;
				for (int j = 0; j < count[1]; j++)
				{
					point[2] = size[2]+1 + size[2]/2 + (size[2]+1)*distance_rel/2;
					for (int k = 0; k < count[2]; k++)
					{
						setStruct(array, dim, point, step, 2);
						point[2] += step[2];
					}
					point[1] += step[1];
				}
				point[0] += step[0];
			}
		}
	}
}


bool VolumeRayCastCompositeFunction::setRegion(char *array, int dim [3], int point [3], float size [3], int value)
{
	if (!isInside(dim, point))
		return false;
	int right [3];
	right[0] = floor(size[0]/2);
	right[1] = floor(size[1]/2);
	right[2] = floor(size[2]/2);
	int left [3];
	left[0] = size[0] - right[0];
	left[1] = size[1] - right[1];
	left[2] = size[2] - right[2];
	for (int i = point[0] - left[0]; i <= point[0] + right[0]; i++)
	{
		for (int j = point[1] - left[1]; j <= point[1] + right[1]; j++)
		{
			for (int k = point[2] - left[2]; k <= point[2] + right[2]; k++)
			{
				int help [3] = {i, j, k};
				if (isInside(dim, help))
				{
					if (array[dim[1]*dim[0]*k + dim[0]*j + i] == 0)
						array[dim[1]*dim[0]*k + dim[0]*j + i] = value;
				}
			}
		}
	}
	return true;
}


bool VolumeRayCastCompositeFunction::setStruct(char *array, int dim [3], int point [3], float step [3], int value)
{
	if (!isInside(dim, point))
		return false;
	int right [3];
	right[0] = floor(step[0]/2);
	right[1] = floor(step[1]/2);
	right[2] = floor(step[2]/2);
	int left [3];
	left[0] = step[0] - right[0];
	left[1] = step[1] - right[1];
	left[2] = step[2] - right[2];
	bool flag;
	for (int i = point[0] - left[0]; i <= point[0] + right[0]; i++)
	{
		int help [3] = {i, point[1], point[2]};
		flag = isInside(dim, help);

		if (flag)
		{
			if (array[dim[1]*dim[0]*point[2] + dim[0]*point[1] + i] == 0)
				array[dim[1]*dim[0]*point[2] + dim[0]*point[1] + i] = value;
		}
	}
	for (int j = point[1] - left[1]; j <= point[1] + right[1]; j++)
	{
		int help [3] = {point[0], j, point[2]};
		flag = isInside(dim, help);

		if (flag)
		{
			if (array[dim[1]*dim[0]*point[2] + dim[0]*j + point[0]] == 0)
				array[dim[1]*dim[0]*point[2] + dim[0]*j + point[0]] = value;
		}
	}
	for (int k = point[2] - left[2]; k <= point[2] + right[2]; k++)
	{
		int help [3] = {point[0], point[1], k};
		flag = isInside(dim, help);

		if (flag)
		{
			if (array[dim[1]*dim[0]*k + dim[0]*point[1] + point[0]] == 0)
				array[dim[1]*dim[0]*k + dim[0]*point[1] + point[0]] = value;
		}
	}
	
	return true;
}


bool VolumeRayCastCompositeFunction::isInside(int dim [3], int point [3])
{
	if (point[0] < 0 || point[1] < 0 || point[2] < 0)
		return false;
	if (point[0] >= dim[0] || point[1] >= dim[1] || point[2] >= dim[2])
		return false;
	return true;
}


void VolumeRayCastCompositeFunction::setDitheringEnable()
{
	if (dithering_pattern_loaded)
		datas.DITHERING = true;
	else
		vtkWarningMacro ( << "No Volume for context is being set." );
}


void VolumeRayCastCompositeFunction::setDitheringDisable()
{
	datas.DITHERING = false;
}


bool VolumeRayCastCompositeFunction::getDitheringStatus()
{
	return datas.DITHERING;
}


void VolumeRayCastCompositeFunction::setMagicLamp(float direction [3], float angle, float transfer_factor, bool backside)
{
	datas.ml_direction[0] = direction[0];
	datas.ml_direction[1] = direction[1];
	datas.ml_direction[2] = direction[2];
	datas.ml_angle = cos(angle*(3.14159265)/180);
	datas.ml_transfer = transfer_factor;
	datas.ml_back_front = backside;
}


void VolumeRayCastCompositeFunction::setMagicLampDirection(float direction [3])
{
	datas.ml_direction[0] = direction[0];
	datas.ml_direction[1] = direction[1];
	datas.ml_direction[2] = direction[2];
}


float * VolumeRayCastCompositeFunction::getMagicLampDirection()
{
	return datas.ml_direction;
}


void VolumeRayCastCompositeFunction::setMagicLampEnable()
{
	datas.MAGIC_LAMP = true;
}


void VolumeRayCastCompositeFunction::setMagicLampDisable()
{
	datas.MAGIC_LAMP = false;
}


bool VolumeRayCastCompositeFunction::getMagicLampStatus()
{
	return datas.MAGIC_LAMP;
}


// This is called from RenderAnImage (in vtkDepthPARCMapper.cxx)
// It uses the integer data type flag that is passed in to
// determine what type of ray needs to be cast (which is handled
// by a templated function.  It also uses the shading and
// interpolation types to determine which templated function
// to call.
void VolumeRayCastCompositeFunction::CastRay( vtkVolumeRayCastDynamicInfo *dynamicInfo,
                                                 vtkVolumeRayCastStaticInfo *staticInfo )
{
  void *data_ptr;

  data_ptr = staticInfo->ScalarDataPointer;

  // Cast the ray for the data type and shading/interpolation type
  if ( staticInfo->InterpolationType == VTK_NEAREST_INTERPOLATION )
    {
    if ( staticInfo->Shading == 0 )
      {
      // Nearest neighbor and no shading
      switch ( staticInfo->ScalarDataType )
        {
        case VTK_UNSIGNED_CHAR:
          vtkCastRay_NN_Unshaded( (unsigned char *)data_ptr, dynamicInfo, 
                                  staticInfo );
          break;
        case VTK_UNSIGNED_SHORT:
          vtkCastRay_NN_Unshaded( (unsigned short *)data_ptr, dynamicInfo, 
                                  staticInfo );
          break;
        default:
          vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
          break;
        }
      }
    else
      {
      // Nearest neighbor and shading
      switch ( staticInfo->ScalarDataType )
        {
        case VTK_UNSIGNED_CHAR:
          vtkCastRay_NN_Shaded( (unsigned char *)data_ptr, dynamicInfo, staticInfo);
          break;
        case VTK_UNSIGNED_SHORT:
          vtkCastRay_NN_Shaded( (unsigned short *)data_ptr, dynamicInfo, staticInfo);
          break;
        default:
          vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
          break;
        }
      }
    }
  else 
    {
    // Trilinear interpolation and no shading
    if ( staticInfo->Shading == 0 )
      {
      if ( this->CompositeMethod == VTK_COMPOSITE_INTERPOLATE_FIRST )
        {
        switch ( staticInfo->ScalarDataType )
          {
          case VTK_UNSIGNED_CHAR:
            vtkCastRay_TrilinSample_Unshaded( (unsigned char *)data_ptr,  
                                              dynamicInfo, staticInfo, datas, ditheringPattern);
            break;
          case VTK_UNSIGNED_SHORT:
            vtkCastRay_TrilinSample_Unshaded( (unsigned short *)data_ptr, 
                                              dynamicInfo, staticInfo, datas, ditheringPattern);
            break;
          default:
            vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
            break;
          }
        }
      else
        {
        switch ( staticInfo->ScalarDataType )
          {
          case VTK_UNSIGNED_CHAR:
            vtkCastRay_TrilinVertices_Unshaded( (unsigned char *)data_ptr,  
                                                dynamicInfo, staticInfo );
            break;
          case VTK_UNSIGNED_SHORT:
            vtkCastRay_TrilinVertices_Unshaded( (unsigned short *)data_ptr, 
                                                dynamicInfo, staticInfo );
            break;
          default:
            vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
            break;
          }
        }
      } 
    else
      {
      // Trilinear interpolation and shading
      if ( this->CompositeMethod == VTK_COMPOSITE_INTERPOLATE_FIRST )
        {
        switch ( staticInfo->ScalarDataType )
          {
          case VTK_UNSIGNED_CHAR:
            vtkCastRay_TrilinSample_Shaded( (unsigned char *)data_ptr, 
                                            dynamicInfo, staticInfo, datas, ditheringPattern);
            break;
          case VTK_UNSIGNED_SHORT:
            vtkCastRay_TrilinSample_Shaded( (unsigned short *)data_ptr, 
                                            dynamicInfo, staticInfo, datas, ditheringPattern);
            break;
          default:
            vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
            break;
          }
        }       
      else
        {
        switch ( staticInfo->ScalarDataType )
          {
          case VTK_UNSIGNED_CHAR:
            vtkCastRay_TrilinVertices_Shaded( (unsigned char *)data_ptr, 
                                              dynamicInfo, staticInfo );
            break;
          case VTK_UNSIGNED_SHORT:
            vtkCastRay_TrilinVertices_Shaded( (unsigned short *)data_ptr, 
                                              dynamicInfo, staticInfo );
            break;
          default:
            vtkWarningMacro ( << "Unsigned char and unsigned short are the only supported datatypes for rendering" );
            break;
          }
        }
      }
    }
}

float VolumeRayCastCompositeFunction::GetZeroOpacityThreshold( vtkVolume 
                                                                  *vol )
{
  return vol->GetProperty()->GetScalarOpacity()->GetFirstNonZeroValue();
}

// We don't need to do any specific initialization here...
void VolumeRayCastCompositeFunction::SpecificFunctionInitialize( 
                                vtkRenderer *vtkNotUsed(ren), 
                                vtkVolume *vtkNotUsed(vol),
                                vtkVolumeRayCastStaticInfo *vtkNotUsed(staticInfo),
                                vtkVolumeRayCastMapper *vtkNotUsed(mapper) )
{
}

// Description:
// Return the composite method as a descriptive character string.
const char *VolumeRayCastCompositeFunction::GetCompositeMethodAsString(void)
{
  if( this->CompositeMethod == VTK_COMPOSITE_INTERPOLATE_FIRST )
    {
    return "Interpolate First";
    }
  if( this->CompositeMethod == VTK_COMPOSITE_CLASSIFY_FIRST )
    {
    return "Classify First";
    }
  else
    {
    return "Unknown";
    }
}

// Print method for vtkVolumeRayCastCompositeFunction
// Since there is nothing local to print, just print the object stuff.
void VolumeRayCastCompositeFunction::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);

  os << indent << "Composite Method: " << this->GetCompositeMethodAsString()
     << "\n";

}





